/* This file is a part of the program Kira.
 * Copyright (C) Johann Usovitsch <jusovitsch@googlemail.com>
 * Philipp Maierhoefer <particle@maierhoefer.net>
 * Peter Uwer <peter.uwer@physik.hu-berlin.de>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version , or (at
 * your option) any later version as published by the Free Software 
 * Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
*/
#ifndef PYRED_INTEGRALS_H
#define PYRED_INTEGRALS_H

#include <cstddef>
#include <iostream>
#include <sstream>
#include <string>
#include <array>
#include <vector>
#include <unordered_set>
#include <unordered_map>
#include <memory>
#include <iterator>
#include <functional> // hash
#include <mutex>
#include <exception>

#include "defs.h"

namespace pyred {

/******************
 * Seed generator *
 ******************/

using compos_it_type = std::vector<std::vector<uint32_t>>::const_iterator;
using compos_rit_type = std::vector<std::vector<uint32_t>>::const_reverse_iterator;

class SeedIterator;

class SeedSpec {
public:
  uint32_t sector;
  uint32_t dots;
  uint32_t sps;
  bool dots_means_r;
};

class Seeds {
public:
  typedef SeedIterator const_iterator;
  Seeds(const Seeds&) = default;
  Seeds(Seeds&&) = default;
  Seeds(uint32_t, uint32_t, uint32_t, uint32_t);
  SeedIterator begin() const;
  SeedIterator cbegin() const;
  SeedIterator end() const;
  SeedIterator cend() const;
private:
  std::vector<bool> sector;
  compos_rit_type ppows_begin, ppows_end;
  compos_it_type spows_begin, spows_end;
};

class SeedIterator {
public:
  typedef std::ptrdiff_t difference_type;
  typedef std::vector<pow_type> value_type;
  typedef const std::vector<pow_type>* pointer;
  typedef const std::vector<pow_type>& reference;
  typedef std::input_iterator_tag iterator_category;
  SeedIterator();
  SeedIterator(const SeedIterator&) = default;
  SeedIterator(SeedIterator&&) = default;
  SeedIterator(const std::vector<bool>&,
               const compos_rit_type&, const compos_rit_type&,
               const compos_it_type&, const compos_it_type&);
  SeedIterator& operator=(const SeedIterator&) = default;
  SeedIterator& operator++();
  SeedIterator operator++(int);
  std::vector<pow_type> operator*() const;
  std::unique_ptr<std::vector<pow_type>> operator->() const;
  friend bool operator==(const SeedIterator&, const SeedIterator&);
  friend void swap(SeedIterator&, SeedIterator&);
private:
  std::vector<bool> sector;
  compos_rit_type p_begin, p_it, p_end;
  compos_it_type s_it, s_end;
};

inline SeedIterator Seeds::begin() const {
  return SeedIterator(sector, ppows_begin, ppows_end, spows_begin, spows_end);
}

inline SeedIterator Seeds::end() const {
  return SeedIterator();
}

inline SeedIterator Seeds::cbegin() const {
  return begin();
}

inline SeedIterator Seeds::cend() const {
  return end();
}

bool operator==(const SeedIterator&, const SeedIterator&);
bool operator!=(const SeedIterator&, const SeedIterator&);
void swap(SeedIterator&, SeedIterator&);

/******************
 * Integral class *
 ******************/

class IntegralRelations;
class Topology;

class IntegralProperties {
public:
  uint32_t topology;
  uint32_t sector;
  uint32_t lines;
  uint32_t nums;
  uint32_t dots;
  uint32_t sps;
  uint32_t ppows_weight;
  uint32_t spows_weight;
};

class Integral {
  friend class IntegralRelations;
  friend class Topology;
  friend bool operator==(const Integral&, const Integral&);
  friend bool operator!=(const Integral&, const Integral&);
public:
  // constructors
  Integral() = default;
  Integral(const Integral&) = default;
  Integral(Integral&&) = default;
  Integral(uint32_t, const std::vector<pow_type>&);
  Integral(uint32_t, std::vector<pow_type>&&);
  Integral(std::istringstream&, uint32_t);
  Integral(weight_type);
  // members
  uint32_t m_topoid;
  std::vector<pow_type> m_powers;
  // methods
  weight_type to_weight() const;
  void reserve(std::size_t);
  std::size_t size() const;
  Integral zip(const std::vector<pow_type>&) const;
  Integral& operator=(const Integral&) = default;
  bool operator<(const Integral&) const;
  // static methods
  static void setup(uint32_t = 1, const std::vector<uint32_t>& = {});
  static void assign_weight_bits();
  static uint32_t np() {return s_np;}
  static void use_cache(int=3);
  static void clear_cache(int=3);
  static IntegralProperties properties(weight_type);
  static Seeds seeds(uint32_t, uint32_t, uint32_t);
private:
  // static methods
  static std::pair<uint32_t,uint32_t> weight_to_dots(
    uint32_t, uint32_t, uint32_t);
  static void weight_overflow(
    uint32_t, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t);
  static uint32_t s_np;
  static uint32_t s_ordering;
  static uint32_t s_minlines;
  // static members
  // s_trivialsectors and s_sector_weight per topology in vectors
  // are duplicated here (wrt. the Topology objects) for faster access
  // than via Topology::s_id_to_topo(m_topoid)
  static std::vector<std::unordered_set<uint32_t>> s_trivialsectors;
  static std::vector<std::vector<std::pair<uint32_t,uint32_t>>> s_sector_weight;
  // topology | sector | dots+sps | dots | ppows | spows || sum
  //     rest |   s_np |        5 |    4 |    17 |    13 ||  64
  static std::array<uint32_t,6> s_weight_bits;
  static std::array<weight_type,6> s_weight_proj;
  // cache_level: 0=(no cache), 1=(weight to integral only),
  //              2=(integral to weight only), 3=(both)
  static int s_cache_level;
  static std::unordered_map<Integral,weight_type> s_i2w_cache;
  static std::unordered_map<weight_type,Integral> s_w2i_cache;
  static std::mutex s_i2w_mtx;
  static std::mutex s_w2i_mtx;
};


inline void Integral::reserve(std::size_t sz) {
  m_powers.reserve(sz);
}

inline std::size_t Integral::size() const {
  return m_powers.size();
}

std::ostream& operator<<(std::ostream&, const Integral&);

/******************
 * Topology class *
 ******************/

class Topology {
  // All non-static public members of this class must be const.
  // Otherwise a copy of a topology may be modified
  // and therfore not agree with the one stored in s_id_to_topo.
  // TODO: can this problem also be solved by deleting the copy
  // constructor and the copy assigment operator?
  // NOTE: Integrals must only be created after all topologies
  //       have been created. The reason for this is that Integral::s_minlines
  //       is static and global for all topologies, i.e. it might change
  //       if new topologies are defined.
  friend class Integral;
  friend Topology& new_topology(const std::string&, const uint32_t,
                                const std::vector<uint32_t>&);
public:
  const std::string m_name;
  const uint32_t m_np;
  const uint32_t m_id;
  const std::unordered_set<uint32_t> m_trivialsectors;
  Topology() = default;
  Integral operator()(const std::vector<pow_type>&);
  Integral operator()(std::vector<pow_type>&&);
  static Topology& id_to_topo(uint32_t topoid) {return *s_id_to_topo.at(topoid);}
private:
  Topology(const std::string&, uint32_t,
           const std::vector<uint32_t>&);
  void assign_sector_weight();
  uint32_t sector_to_subweight(uint32_t s) {
    return m_sector_weight[s].first;
  }
  uint32_t sector_from_subweight(uint32_t w) {
    return m_sector_weight[w].second;
  }
  std::vector<std::pair<uint32_t,uint32_t>> m_sector_weight;
  uint32_t m_minlines;
  static uint32_t s_next_id;
  // use a pointer here so that references don't get invalidated
  static std::vector<std::unique_ptr<Topology>> s_id_to_topo;
};

inline Integral Topology::operator()(const std::vector<pow_type>& pows) {
  return Integral(m_id, pows);
}

inline Integral Topology::operator()(std::vector<pow_type>&& pows) {
  return Integral(m_id, std::move(pows));
}

// The user will probably make a copy of the topology from s_id_to_topo,
// but that's ok, because all members are const.
Topology& new_topology(const std::string&,
                       const uint32_t, const std::vector<uint32_t>&);
Topology& new_topology(const std::string&, const std::string&);

}


namespace std {
  template<>
  struct hash<pyred::Integral> {
    size_t operator()(const pyred::Integral& igl) const {
      size_t h = igl.m_topoid + 0x9e3779b9;
      for (const auto& pow: igl.m_powers) {
        h ^= pow + 0x9e3779b9 + (h << 6) + (h >> 2);
      }
      return h;
    }
  };
}

#endif
